"""tags 子命令"""

import json

from .. import warn


def add_parser(sub):
    p = sub.add_parser("tags", help="列出所有标签")
    p.set_defaults(func=cmd_tags)
    return p


def cmd_tags(conn, args):
    rows = conn.execute("SELECT tags FROM memos WHERE tags IS NOT NULL AND tags != ''").fetchall()
    tags = set()
    for row in rows:
        try:
            tags.update(json.loads(row["tags"]))
        except json.JSONDecodeError:
            warn(f"skipped malformed tags: {row['tags']}")

    result = sorted(tags)
    print(json.dumps(result, ensure_ascii=False, indent=2))



"""add 子命令"""

import json
import time

from .. import render_memo_text, split_tags_and_content


def add_parser(sub):
    p = sub.add_parser("add", help="添加 memo")
    p.add_argument("content", nargs="+")
    p.set_defaults(func=cmd_add)
    return p


def cmd_add(conn, args):
    raw_content = " ".join(args.content).strip()
    if not raw_content:
        raise ValueError("content required")

    tags, content = split_tags_and_content(raw_content)
    if not content:
        raise ValueError("content required")
    now = time.time()

    try:
        with conn:
            cur = conn.execute(
                "INSERT INTO memos (content, tags, created_at, source) VALUES (?, ?, ?, 'cli')",
                (content, json.dumps(tags, ensure_ascii=False), now),
            )
            conn.execute(
                "INSERT INTO memos_fts(rowid, content) VALUES (?, ?)",
                (cur.lastrowid, render_memo_text(tags, content)),
            )
    except Exception as exc:
        raise RuntimeError(f"failed to add: {exc}") from exc

    print(f"created #{cur.lastrowid}")
    if tags:
        print(f"tags: {' '.join('#' + t.lstrip('#') for t in tags)}")



"""共用的格式化、解析、渲染工具函数"""

import random
import re
import unicodedata
from datetime import datetime, timedelta, timezone

TZ_BJ = timezone(timedelta(hours=8))
TAG_PATTERN = r"#[\w一-鿿][\w一-鿿.-]*(?:/[\w一-鿿][\w一-鿿.-]*)*"

KAMI_PARCHMENT = "#f5f4ed"
KAMI_IVORY = "#faf9f5"
KAMI_INK = "#1B365D"
KAMI_TAG_BG = "#E4ECF5"
KAMI_NEAR_BLACK = "#141413"
KAMI_OLIVE = "#504e49"
KAMI_STONE = "#6b6a64"
KAMI_BORDER = "#e8e6dc"

RELATION_TYPES = {"related", "supports", "contrasts"}
REVIEW_LAYERS = (
    ("fresh", 0.20),
    ("middle", 0.45),
    ("old", 0.35),
)
REVIEW_FALLBACK_LAYERS = ("old", "middle", "fresh")


# ── tags ──────────────────────────────────────────────────────────────────────


def parse_tags(text):
    tags = re.findall(TAG_PATTERN, text, re.UNICODE)
    return list(dict.fromkeys(t.lower() for t in tags))


def clean_content(text):
    text = re.sub(r"[ \t\r\f\v]+", " ", text)
    text = re.sub(r" *\n *", "\n", text)
    text = re.sub(r"\n{3,}", "\n\n", text)
    return text.strip()


def remove_tags(text):
    return clean_content(re.sub(TAG_PATTERN, "", text, flags=re.UNICODE))


def split_tags_and_content(text):
    return parse_tags(text), remove_tags(text)


def render_memo_text(tags, content):
    tag_str = " ".join(f"#{tag.lstrip('#')}" for tag in tags)
    if tag_str and content:
        return f"{tag_str} {content}"
    return tag_str or content


# ── time ──────────────────────────────────────────────────────────────────────


def fmt_time(ts):
    dt = datetime.fromtimestamp(ts, TZ_BJ)
    return f"{dt.year}/{dt.month:02d}/{dt.day:02d} {dt.hour:02d}:{dt.minute:02d}:{dt.second:02d}"


def fmt_datetime(ts):
    return fmt_time(ts)


def fmt_review_time(ts):
    return fmt_time(ts)


def fmt_history_time(ts):
    return fmt_time(ts)


def history_file_path(ts, history_dir):
    dt = datetime.fromtimestamp(ts, TZ_BJ)
    return history_dir / f"{dt.year}-{dt.month:02d}-{dt.day:02d}.md"


# ── text wrap ─────────────────────────────────────────────────────────────────


def text_units(text):
    total = 0
    for char in text:
        total += 2 if unicodedata.east_asian_width(char) in {"F", "W"} else 1
    return total


def wrap_line(line, max_units):
    if not line:
        return [""]
    lines = []
    current = ""
    current_units = 0
    for char in line:
        unit = 2 if unicodedata.east_asian_width(char) in {"F", "W"} else 1
        if current and current_units + unit > max_units:
            lines.append(current)
            current = char
            current_units = unit
        else:
            current += char
            current_units += unit
    if current:
        lines.append(current)
    return lines


def wrap_text(text, max_units):
    lines = []
    for paragraph in text.splitlines():
        lines.extend(wrap_line(paragraph, max_units))
    return lines


# ── review ─────────────────────────────────────────────────────────────────────


def choose_review_layer():
    roll = random.random()
    cursor = 0
    for layer, probability in REVIEW_LAYERS:
        cursor += probability
        if roll < cursor:
            return layer
    return REVIEW_LAYERS[-1][0]


def review_layer_bounds(layer):
    if layer == "fresh":
        return 7, 30
    if layer == "middle":
        return 30, 180
    if layer == "old":
        return 180, None
    raise ValueError(f"unknown review layer: {layer}")


# ── search ────────────────────────────────────────────────────────────────────


def fts_query(text):
    tokens = re.findall(r"[\w一-鿿]+", text.lower(), re.UNICODE)
    escaped = [token.replace('"', '""') for token in tokens]
    return " ".join(f'"{token}"*' for token in escaped if token)


# ── link ──────────────────────────────────────────────────────────────────────


def normalize_link_ids(left_id, right_id):
    if left_id == right_id:
        raise ValueError("cannot link a memo to itself")
    return (left_id, right_id) if left_id < right_id else (right_id, left_id)


def validate_relation_type(value):
    relation_type = value.lower()
    if relation_type not in RELATION_TYPES:
        allowed = ", ".join(sorted(RELATION_TYPES))
        raise ValueError(f"relation type must be one of: {allowed}")
    return relation_type



"""backup 子命令"""

from pathlib import Path

from .. import create_backup


def add_parser(sub):
    p = sub.add_parser("backup", help="备份数据库")
    p.add_argument("--out", help="输出 .db 文件路径")
    p.set_defaults(func=cmd_backup)
    return p


def cmd_backup(conn, args):
    out = create_backup(conn, Path(args.out) if args.out else None)
    print(f"backup saved: {out}")



"""reset 子命令 — 重置数据目录"""

import shutil
import time
from pathlib import Path

from .. import DEFAULT_DATA_DIR, create_backup, get_data_dir, get_db_path


def add_parser(sub):
    p = sub.add_parser("reset", help="重置数据目录")
    p.add_argument("--force", action="store_true", help="skip confirmation")
    p.add_argument("--no-backup", action="store_true", help="skip automatic backup")
    p.set_defaults(func=cmd_reset)
    return p


def _safe_reset_path(path):
    resolved = path.expanduser().resolve()
    home = Path.home().resolve()
    if resolved in {Path("/"), home, home.parent}:
        raise ValueError(f"refusing to reset unsafe data dir: {resolved}")
    if resolved == DEFAULT_DATA_DIR.resolve():
        return resolved
    if (resolved / "memo.db").exists():
        return resolved
    if "memo" not in resolved.name.lower():
        raise ValueError(f"refusing to reset data dir without memo in its name: {resolved}")
    return resolved


def _reset_backup_path(data_dir):
    stamp = time.strftime("%Y%m%d-%H%M%S")
    return data_dir.parent / f"{data_dir.name}-backup-{stamp}.db"


def cmd_reset(conn, args):
    data_dir = _safe_reset_path(get_data_dir())
    if not data_dir.exists():
        print(f"data dir not found: {data_dir}")
        return 0

    if not args.force:
        print(f"will delete data dir: {data_dir}")
        print("includes: memo.db, images/, history/")
        print()
        confirm = input("type 'yes' to confirm: ")
        if confirm.strip() != "yes":
            print("cancelled")
            return 0

    backup_path = None
    if not args.no_backup and get_db_path().exists():
        backup_path = create_backup(conn, _reset_backup_path(data_dir))

    shutil.rmtree(data_dir)
    if backup_path:
        print(f"backup saved: {backup_path}")
    print(f"deleted: {data_dir}")
    return 0



"""rebuild_fts 子命令"""

from .. import rebuild_fts_index


def add_parser(sub):
    p = sub.add_parser("rebuild-fts", help="重建搜索索引")
    p.set_defaults(func=cmd_rebuild_fts)
    return p


def cmd_rebuild_fts(conn, args):
    try:
        with conn:
            count = rebuild_fts_index(conn)
    except Exception as exc:
        raise RuntimeError(f"failed to rebuild fts: {exc}") from exc

    print(f"fts rebuilt ({count} memos indexed)")



"""Render memo share images as PNG."""

import json
from pathlib import Path

from .helpers import (
    KAMI_INK,
    KAMI_IVORY,
    KAMI_NEAR_BLACK,
    KAMI_PARCHMENT,
    KAMI_STONE,
    fmt_datetime,
    text_units,
)


def _load_pillow():
    from PIL import Image, ImageDraw, ImageFont

    return Image, ImageDraw, ImageFont


def _font_paths(weight):
    bundled_fonts = Path(__file__).resolve().parent / "assets" / "fonts"
    home_fonts = Path.home() / "Library" / "Fonts"
    if weight == "medium":
        return [
            bundled_fonts / "LXGWWenKai-Medium.ttf",
            home_fonts / "LXGWWenKai-Medium.ttf",
            home_fonts / "LXGWWenKaiMono-Medium.ttf",
            "/System/Library/Fonts/Hiragino Sans GB.ttc",
            "/System/Library/Fonts/STHeiti Medium.ttc",
            "/usr/share/fonts/opentype/noto/NotoSansCJK-Regular.ttc",
            "/usr/share/fonts/opentype/noto/NotoSerifCJK-Regular.ttc",
            "/usr/share/fonts/truetype/noto/NotoSansCJK-Regular.ttc",
            "/usr/share/fonts/truetype/noto/NotoSerifCJK-Regular.ttc",
            "/usr/share/fonts/truetype/dejavu/DejaVuSans.ttf",
        ]
    return [
        bundled_fonts / "LXGWWenKai-Regular.ttf",
        home_fonts / "LXGWWenKai-Regular.ttf",
        home_fonts / "LXGWWenKaiMono-Regular.ttf",
        "/System/Library/Fonts/Hiragino Sans GB.ttc",
        "/System/Library/Fonts/STHeiti Light.ttc",
        "/System/Library/Fonts/Supplemental/Songti.ttc",
        "/usr/share/fonts/opentype/noto/NotoSansCJK-Regular.ttc",
        "/usr/share/fonts/opentype/noto/NotoSerifCJK-Regular.ttc",
        "/usr/share/fonts/truetype/noto/NotoSansCJK-Regular.ttc",
        "/usr/share/fonts/truetype/noto/NotoSerifCJK-Regular.ttc",
        "/usr/share/fonts/truetype/dejavu/DejaVuSans.ttf",
    ]


def _font(font_module, size, weight="regular"):
    for path in _font_paths(weight):
        candidate = Path(path)
        if candidate.exists():
            return font_module.truetype(str(candidate), size=size)
    return font_module.load_default(size=size)


def _text_size(draw, text, font):
    bbox = draw.textbbox((0, 0), text, font=font)
    return bbox[2] - bbox[0], bbox[3] - bbox[1]


def _draw_text_top(draw, xy, text, font, fill):
    x, top = xy
    bbox = draw.textbbox((0, 0), text, font=font)
    draw.text((x - bbox[0], top - bbox[1]), text, fill=fill, font=font)


def _draw_text_center_y(draw, xy, text, font, fill, box_height):
    x, top = xy
    _, text_height = _text_size(draw, text, font)
    _draw_text_top(draw, (x, top + (box_height - text_height) / 2), text, font, fill)


def _wrap_text_pixels(draw, text, font, max_width):
    leading_punctuation = "，。；：？！、,.!?;:)]}）】》"
    lines = []
    for paragraph in text.splitlines():
        if not paragraph:
            lines.append("")
            continue
        current = ""
        for char in paragraph:
            candidate = current + char
            if current and draw.textlength(candidate, font=font) > max_width:
                if char in leading_punctuation:
                    lines.append(candidate)
                    current = ""
                else:
                    lines.append(current)
                    current = char
            else:
                current = candidate
        if current:
            lines.append(current)
    return lines


def _chip_width(draw, tag, font, scale):
    return int(draw.textlength(tag, font=font) / scale + 20)


def _chip_rows(draw, tags, width, margin_x, font, scale):
    rows = []
    current = []
    current_width = 0
    max_width = width - margin_x * 2
    for tag in tags:
        chip_width = _chip_width(draw, tag, font, scale)
        if current and current_width + chip_width + 10 > max_width:
            rows.append(current)
            current = [tag]
            current_width = chip_width
        else:
            current.append(tag)
            current_width += chip_width + 10
    if current:
        rows.append(current)
    return rows


def _content_metrics(content):
    units = max((text_units(line) for line in content.splitlines()), default=0)
    total_units = text_units(content)
    if total_units <= 36 and units <= 28:
        size = 30
    elif total_units <= 86 and units <= 38:
        size = 28
    elif total_units <= 180:
        size = 25
    elif total_units <= 360:
        size = 23
    else:
        size = 21
    return size, int(size * 1.72)


def _theme(style):
    themes = {
        "clean": {
            "page": "#f6f6f2",
            "card": "#ffffff",
            "shadow": None,
            "border": None,
            "tag_bg": "#eeeeeb",
            "tag_text": "#5c625f",
            "content": "#151514",
            "footer": "#7b7b75",
            "brand": "#3f4643",
            "divider_color": "#deded8",
            "radius": 0,
            "card_margin": 0,
            "pad_y": 50,
            "content_x": 54,
            "tag_radius": 13,
            "shadow_offset": 0,
            "show_divider": True,
        },
        "paper": {
            "page": KAMI_PARCHMENT,
            "card": KAMI_IVORY,
            "shadow": "#eeece2",
            "border": "#e8e4d8",
            "tag_bg": "#edf3f7",
            "tag_text": KAMI_INK,
            "content": KAMI_NEAR_BLACK,
            "footer": KAMI_STONE,
            "brand": KAMI_INK,
            "divider_color": "#ebe7dc",
            "radius": 18,
            "card_margin": 26,
            "pad_y": 38,
            "content_x": 52,
            "tag_radius": 3,
            "shadow_offset": 4,
            "show_divider": True,
        },
        "ink": {
            "page": "#ebe9df",
            "card": "#fbfaf2",
            "shadow": "#dedace",
            "border": "#233a5d",
            "tag_bg": "#233a5d",
            "tag_text": "#f7f5ec",
            "content": "#111827",
            "footer": "#626863",
            "brand": "#172f52",
            "divider_color": "#233a5d",
            "radius": 10,
            "card_margin": 24,
            "pad_y": 42,
            "content_x": 50,
            "tag_radius": 2,
            "shadow_offset": 6,
            "show_divider": True,
        },
    }
    if style not in themes:
        allowed = ", ".join(themes)
        raise ValueError(f"image style must be one of: {allowed}")
    return themes[style]


def render_png(row, out_path, total_memos, width=600, style="paper"):
    Image, ImageDraw, ImageFont = _load_pillow()

    scale = 3
    canvas_width = width * scale
    out_path = Path(out_path)
    theme = _theme(style)

    tags = json.loads(row["tags"]) if row["tags"] else []
    tag_items = [f"#{tag.lstrip('#')}" for tag in tags] or ["memo"]
    content = row["content"].strip()
    content_size, content_line_h = _content_metrics(content)
    created_at = fmt_datetime(row["created_at"])
    brand_text = f"a-memo · {total_memos} memos"

    card_margin = theme["card_margin"]
    card_pad_y = theme["pad_y"]
    content_x = theme["content_x"]
    content_right = width - theme["content_x"]
    tag_h = 26
    tag_row_gap = 6
    content_gap = 34
    footer_gap = 38
    footer_h = 28
    tag_font = _font(ImageFont, 18 * scale, "medium")
    content_font = _font(ImageFont, content_size * scale, "medium")
    footer_font = _font(ImageFont, 18 * scale)
    footer_brand_font = _font(ImageFont, 18 * scale, "medium")

    def s(value):
        return int(round(value * scale))

    measure = Image.new("RGB", (1, 1), theme["page"])
    measure_draw = ImageDraw.Draw(measure)
    rows = _chip_rows(measure_draw, tag_items, width, content_x, tag_font, scale)
    tag_top = card_margin + card_pad_y
    tags_h = len(rows) * tag_h + max(0, len(rows) - 1) * tag_row_gap
    content_top = tag_top + tags_h + content_gap
    content_lines = _wrap_text_pixels(
        measure_draw,
        content,
        content_font,
        s(content_right - content_x),
    )
    footer_top = content_top + len(content_lines) * content_line_h + footer_gap
    content_height = footer_top + footer_h + card_pad_y - card_margin
    height = card_margin * 2 + content_height
    image = Image.new("RGB", (canvas_width, height * scale), theme["page"])
    draw = ImageDraw.Draw(image)

    shadow_offset = theme["shadow_offset"]
    radius = theme["radius"]
    if theme["shadow"]:
        draw.rounded_rectangle(
            (
                s(card_margin + shadow_offset),
                s(card_margin + shadow_offset),
                s(width - card_margin + shadow_offset),
                s(height - card_margin + shadow_offset),
            ),
            radius=s(radius),
            fill=theme["shadow"],
        )
    draw.rounded_rectangle(
        (s(card_margin), s(card_margin), s(width - card_margin), s(height - card_margin)),
        radius=s(radius),
        fill=theme["card"],
        outline=theme["border"],
        width=s(1) if theme["border"] else 0,
    )

    y = tag_top
    for row_tags in rows:
        x = content_x
        for tag in row_tags:
            chip_width = _chip_width(draw, tag, tag_font, scale)
            draw.rounded_rectangle(
                (s(x), s(y), s(x + chip_width), s(y + tag_h)),
                radius=s(theme["tag_radius"]),
                fill=theme["tag_bg"],
            )
            _draw_text_center_y(draw, (s(x + 10), s(y)), tag, tag_font, theme["tag_text"], s(tag_h))
            x += chip_width + 10
        y += tag_h + tag_row_gap

    y = content_top
    for line in content_lines:
        _draw_text_top(
            draw,
            (s(content_x), s(y)),
            line,
            content_font,
            theme["content"],
        )
        y += content_line_h

    if theme["show_divider"]:
        divider_y = footer_top - 14
        draw.line(
            (s(content_x), s(divider_y), s(content_right), s(divider_y)),
            fill=theme["divider_color"],
            width=s(1),
        )
    _draw_text_center_y(
        draw,
        (s(content_x), s(footer_top)),
        created_at,
        footer_font,
        theme["footer"],
        s(footer_h),
    )
    brand_width = draw.textlength(brand_text, font=footer_brand_font) / scale
    _draw_text_center_y(
        draw,
        (s(content_right - brand_width), s(footer_top)),
        brand_text,
        footer_brand_font,
        theme["brand"],
        s(footer_h),
    )

    image = image.resize((width, height), Image.Resampling.LANCZOS)
    out_path.parent.mkdir(parents=True, exist_ok=True)
    image.save(out_path, "PNG")
    return out_path



"""flomo_import 子命令 — 从 flomo HTML 导出文件导入"""

import json
import re
import sys
import time
from html.parser import HTMLParser
from pathlib import Path
from typing import TypedDict

from .. import render_memo_text, split_tags_and_content


class ParsedMemo(TypedDict):
    time: str
    content: str


class CurrentMemo(TypedDict):
    time: str
    content: list[str]


def add_parser(sub):
    p = sub.add_parser("flomo-import", help="从 flomo HTML 导入")
    p.add_argument("html_path", type=Path)
    p.add_argument("--dry-run", action="store_true")
    p.set_defaults(func=cmd_flomo_import)
    return p


class FlomoHTMLParser(HTMLParser):
    def __init__(self):
        super().__init__(convert_charrefs=True)
        self.memos: list[ParsedMemo] = []
        self.memo_depth = 0
        self.current: CurrentMemo | None = None
        self.field: str | None = None
        self.field_depth = 0

    def handle_starttag(self, tag, attrs):
        attrs = dict(attrs)
        classes = set(attrs.get("class", "").split())

        if tag == "div" and "memo" in classes:
            self.current = {"time": "", "content": []}
            self.memo_depth = 1
            self.field = None
            self.field_depth = 0
            return

        if self.current is None:
            return

        self.memo_depth += 1

        if tag == "div" and "time" in classes:
            self.field = "time"
            self.field_depth = self.memo_depth
        elif tag == "div" and "content" in classes:
            self.field = "content"
            self.field_depth = self.memo_depth
        elif self.field == "content" and tag in {"p", "li", "br"}:
            self._append_content("\n")

    def handle_endtag(self, tag):
        if self.current is None:
            return

        if self.field and self.memo_depth == self.field_depth:
            self.field = None
            self.field_depth = 0

        self.memo_depth -= 1
        if self.memo_depth == 0:
            content = _clean_text("".join(self.current["content"]))
            ts = _clean_text(self.current["time"])
            if ts or content:
                self.memos.append({"time": ts, "content": content})
            self.current = None
            self.field = None
            self.field_depth = 0

    def handle_data(self, data):
        if self.current is None or not self.field:
            return
        if self.field == "time":
            self.current["time"] += data
        elif self.field == "content":
            self._append_content(data)

    def _append_content(self, text):
        if self.current is None:
            return
        self.current["content"].append(text)


def _clean_text(text: str) -> str:
    text = text.replace("\xa0", " ")
    text = re.sub(r"[ \t\r\f\v]+", " ", text)
    text = re.sub(r" *\n *", "\n", text)
    text = re.sub(r"\n{3,}", "\n\n", text)
    return text.strip()


def parse_html(path) -> list[ParsedMemo]:
    parser = FlomoHTMLParser()
    parser.feed(Path(path).read_text(encoding="utf-8"))
    parser.close()
    return parser.memos


def parse_datetime(value):
    try:
        return time.mktime(time.strptime(value.strip(), "%Y-%m-%d %H:%M:%S"))
    except ValueError as exc:
        raise ValueError(f"invalid datetime format: {value}") from exc


def cmd_flomo_import(conn, args):
    if not args.html_path.exists():
        print(f"file not found: {args.html_path}", file=sys.stderr)
        return 1

    try:
        memos = parse_html(args.html_path)
    except OSError as exc:
        print(f"failed to read file: {exc}", file=sys.stderr)
        return 1

    print(f"parsed {len(memos)} memos")

    imported = 0
    skipped = 0
    failed = 0

    if args.dry_run:
        for memo in memos:
            if not memo["content"]:
                skipped += 1
                continue
            tags, content = split_tags_and_content(memo["content"])
            if not content:
                skipped += 1
                continue
            print(
                f"[dry] {memo['time']} | {render_memo_text(tags, content)[:50]}... | tags: {tags}"
            )
            imported += 1
        print(f"\nimported {imported}, skipped {skipped}, failed {failed}")
        print("(dry-run: no data written)")
        return 0

    for memo in memos:
        if not memo["content"]:
            skipped += 1
            continue
        tags, content = split_tags_and_content(memo["content"])
        if not content:
            skipped += 1
            continue
        try:
            ts = parse_datetime(memo["time"])
            with conn:
                cur = conn.execute(
                    """
                    INSERT INTO memos (content, tags, created_at, source)
                    VALUES (?, ?, ?, 'flomo')
                    """,
                    (content, json.dumps(tags, ensure_ascii=False), ts),
                )
                conn.execute(
                    "INSERT INTO memos_fts(rowid, content) VALUES (?, ?)",
                    (cur.lastrowid, render_memo_text(tags, content)),
                )
            imported += 1
        except Exception as exc:
            failed += 1
            print(f"failed: {memo.get('time', '')} | {exc}", file=sys.stderr)

    print(f"\nimported {imported}, skipped {skipped}, failed {failed}")
    return 1 if failed else 0



"""link 子命令"""

import argparse
import time

from .. import clean_content, normalize_link_ids, require_memos, validate_relation_type


def add_parser(sub):
    p = sub.add_parser("link", help="关联两条 memo")
    p.add_argument("left_id", type=positive_int)
    p.add_argument("right_id", type=positive_int)
    p.add_argument("--type", default="related", help="关联类型: related, supports, contrasts")
    p.add_argument("--note", help="关联说明")
    p.set_defaults(func=cmd_link)
    return p


def positive_int(value):
    try:
        parsed = int(value)
    except ValueError as exc:
        raise argparse.ArgumentTypeError("must be a non-negative integer") from exc
    if parsed < 0:
        raise argparse.ArgumentTypeError("must be >= 0")
    return parsed


def cmd_link(conn, args):
    relation_type = validate_relation_type(args.type)
    from_memo_id, to_memo_id = normalize_link_ids(args.left_id, args.right_id)
    require_memos(conn, [from_memo_id, to_memo_id])
    note = clean_content(args.note) if args.note else None

    try:
        with conn:
            cur = conn.execute(
                """
                INSERT OR IGNORE INTO memo_links
                  (from_memo_id, to_memo_id, relation_type, note, created_at, source)
                VALUES (?, ?, ?, ?, ?, 'cli')
                """,
                (from_memo_id, to_memo_id, relation_type, note, time.time()),
            )
            if cur.rowcount == 0 and note:
                conn.execute(
                    """
                    UPDATE memo_links
                    SET note=?
                    WHERE from_memo_id=? AND to_memo_id=? AND relation_type=?
                    """,
                    (note, from_memo_id, to_memo_id, relation_type),
                )
    except Exception as exc:
        raise RuntimeError(f"failed to link: {exc}") from exc

    action = "linked" if cur.rowcount else "link exists"
    note_suffix = f" · {note}" if note else ""
    print(f"{action} #{args.left_id} ↔ #{args.right_id} ({relation_type}){note_suffix}")



import json
import sqlite3
import time

import pytest
from memo import cli
from PIL import Image


@pytest.fixture
def data_dir(tmp_path, monkeypatch):
    path = tmp_path / "memo-data"
    monkeypatch.setenv("MEMO_DATA_DIR", str(path))
    monkeypatch.delenv("MEMO_DB_PATH", raising=False)
    return path


def run_cli(args, capsys):
    code = cli.main(args)
    captured = capsys.readouterr()
    return code, captured.out, captured.err


def parse_json_output(output):
    return json.loads(output)


def db_path(data_dir):
    return data_dir / "memo.db"


def schema_version(data_dir):
    with sqlite3.connect(db_path(data_dir)) as conn:
        return conn.execute("PRAGMA user_version").fetchone()[0]


def make_old_memos(data_dir, count=2):
    created_at = time.time() - 8 * 86400
    with sqlite3.connect(db_path(data_dir)) as conn:
        for memo_id in range(1, count + 1):
            conn.execute(
                "UPDATE memos SET created_at=?, last_review_at=NULL WHERE id=?",
                (created_at - memo_id, memo_id),
            )


def test_help_version_unknown_command_and_invalid_argument(data_dir, capsys):
    with pytest.raises(SystemExit) as help_exit:
        cli.main(["--help"])
    assert help_exit.value.code == 0
    assert "memo <command>" in capsys.readouterr().out

    with pytest.raises(SystemExit) as version_exit:
        cli.main(["--version"])
    assert version_exit.value.code == 0
    assert capsys.readouterr().out.startswith("memo ")

    with pytest.raises(SystemExit) as invalid_arg_exit:
        cli.main(["list", "--limit", "nope"])
    assert invalid_arg_exit.value.code == 2

    with pytest.raises(SystemExit) as unknown_exit:
        cli.main(["llist"])
    assert unknown_exit.value.code == 2
    captured = capsys.readouterr()
    assert "unknown command 'llist'" in captured.err
    assert "did you mean 'list'" in captured.err


def test_init_add_list_search_tags_update_rebuild_and_delete(data_dir, capsys):
    code, out, err = run_cli(["init"], capsys)
    assert (code, out.strip(), err) == (0, "ready", "")
    assert schema_version(data_dir) == 1

    code, out, err = run_cli(["add", "hello", "world", "#Work/AI"], capsys)
    assert code == 0
    assert err == ""
    assert "created #1" in out
    assert "tags: #work/ai" in out

    code, out, _ = run_cli(["list", "#work"], capsys)
    rows = parse_json_output(out)
    assert [row["content"] for row in rows] == ["hello world"]
    assert rows[0]["tags"] == ["#work/ai"]

    code, out, _ = run_cli(["search", "hello"], capsys)
    rows = parse_json_output(out)
    assert code == 0
    assert rows[0]["id"] == 1

    code, out, _ = run_cli(["tags"], capsys)
    assert parse_json_output(out) == ["#work/ai"]

    code, out, err = run_cli(["update", "1", "updated", "memo"], capsys)
    assert code == 0
    assert err == ""
    assert "updated #1" in out
    assert "tags: #work/ai" in out

    code, out, _ = run_cli(["search", "updated"], capsys)
    assert parse_json_output(out)[0]["content"] == "updated memo"

    code, out, err = run_cli(["rebuild-fts"], capsys)
    assert (code, out.strip(), err) == (0, "fts rebuilt (1 memos indexed)", "")

    code, out, err = run_cli(["delete", "1"], capsys)
    assert (code, out.strip(), err) == (0, "deleted #1", "")

    code, out, _ = run_cli(["list"], capsys)
    assert parse_json_output(out) == []


def test_tag_link_links_unlink_and_related_errors(data_dir, capsys):
    run_cli(["add", "first", "#a"], capsys)
    run_cli(["add", "second", "#b"], capsys)

    code, out, err = run_cli(["tag", "1", "#extra"], capsys)
    assert code == 0
    assert err == ""
    assert "tagged #1: #a #extra" in out

    code, out, err = run_cli(
        ["link", "1", "2", "--type", "supports", "--note", "good link"], capsys
    )
    assert code == 0
    assert err == ""
    assert "linked #1 ↔ #2 (supports) · good link" in out

    code, out, _ = run_cli(["links", "1"], capsys)
    links = parse_json_output(out)
    assert links[0]["id"] == 2
    assert links[0]["relation_type"] == "supports"
    assert links[0]["note"] == "good link"

    code, out, err = run_cli(["unlink", "2", "1", "--type", "supports"], capsys)
    assert code == 0
    assert err == ""
    assert "unlinked #2 ↔ #1 (supports)" in out

    code, out, err = run_cli(["unlink", "1", "2", "--type", "supports"], capsys)
    assert code == 1
    assert out == ""
    assert "#1 ↔ #2 has no supports link" in err

    code, _, err = run_cli(["link", "1", "1"], capsys)
    assert code == 1
    assert "cannot link a memo to itself" in err

    code, _, err = run_cli(["links", "99"], capsys)
    assert code == 1
    assert "memo not found: #99" in err


def test_review_push_outputs_links_updates_counts_and_history(data_dir, capsys):
    run_cli(["add", "review one", "#r"], capsys)
    run_cli(["add", "review two", "#r"], capsys)
    run_cli(["link", "1", "2"], capsys)
    make_old_memos(data_dir, count=2)

    code, out, err = run_cli(["review", "--count", "2", "--push"], capsys)
    assert code == 0
    assert err == ""
    rows = parse_json_output(out)
    assert {row["id"] for row in rows} == {1, 2}
    assert any(row["links"] for row in rows)

    code, out, _ = run_cli(["list"], capsys)
    listed = parse_json_output(out)
    assert {row["review_count"] for row in listed} == {1}
    assert all("last_review_at" in row for row in listed)

    history_files = list((data_dir / "history").glob("*.md"))
    assert len(history_files) == 1
    history = history_files[0].read_text(encoding="utf-8")
    assert "#r review one" in history
    assert "#r review two" in history


def test_review_empty_when_no_eligible_memos(data_dir, capsys):
    run_cli(["add", "fresh"], capsys)

    code, out, err = run_cli(["review", "--count", "2"], capsys)
    assert (code, out.strip(), err) == (0, "[]", "")


def test_image_png_rendering(data_dir, tmp_path, capsys):
    run_cli(["add", "shareable memo", "#img"], capsys)

    png_path = tmp_path / "share.png"
    code, out, err = run_cli(["image", "1", "--out", str(png_path)], capsys)
    assert code == 0
    assert err == ""
    assert out.strip() == f"saved: {png_path}"
    assert png_path.read_bytes().startswith(b"\x89PNG\r\n\x1a\n")
    with Image.open(png_path) as image:
        assert image.size[0] == 600
        assert image.size[1] > 0

    extensionless_path = tmp_path / "share-extensionless"
    code, out, err = run_cli(["image", "1", "--out", str(extensionless_path)], capsys)
    assert code == 0
    assert err == ""
    assert out.strip() == f"saved: {extensionless_path.with_suffix('.png')}"
    assert extensionless_path.with_suffix(".png").exists()

    code, _, err = run_cli(["image", "1", "--out", str(tmp_path / "bad.svg")], capsys)
    assert code == 1
    assert "image output requires .png extension" in err

    styled_path = tmp_path / "share-ink.png"
    code, out, err = run_cli(["image", "1", "--style", "ink", "--out", str(styled_path)], capsys)
    assert code == 0
    assert err == ""
    assert out.strip() == f"saved: {styled_path}"
    assert styled_path.exists()

    with pytest.raises(SystemExit) as invalid_style_exit:
        cli.main(["image", "1", "--style", "loud"])
    assert invalid_style_exit.value.code == 2


def test_flomo_import_dry_run_success_and_failures(data_dir, tmp_path, capsys):
    html_path = tmp_path / "flomo.html"
    html_path.write_text(
        """
        <div class="memo">
          <div class="time">2024-01-02 03:04:05</div>
          <div class="content"><p>#flomo imported memo</p><p>second line</p></div>
        </div>
        <div class="memo">
          <div class="time">bad-date</div>
          <div class="content">will fail</div>
        </div>
        """,
        encoding="utf-8",
    )

    code, out, err = run_cli(["flomo-import", str(html_path), "--dry-run"], capsys)
    assert code == 0
    assert err == ""
    assert "parsed 2 memos" in out
    assert "imported 2, skipped 0, failed 0" in out

    code, out, err = run_cli(["flomo-import", str(html_path)], capsys)
    assert code == 1
    assert "imported 1, skipped 0, failed 1" in out
    assert "invalid datetime format: bad-date" in err

    code, out, _ = run_cli(["list", "#flomo"], capsys)
    rows = parse_json_output(out)
    assert len(rows) == 1
    assert rows[0]["content"] == "imported memo\nsecond line"

    code, out, err = run_cli(["flomo-import", str(tmp_path / "missing.html")], capsys)
    assert code == 1
    assert out == ""
    assert "file not found:" in err


def test_reset_force_deletes_data_directory(data_dir, capsys):
    run_cli(["add", "temporary"], capsys)
    assert db_path(data_dir).exists()

    code, out, err = run_cli(["reset", "--force"], capsys)
    assert code == 0
    assert err == ""
    assert "backup saved:" in out
    assert f"deleted: {data_dir}" in out
    assert not data_dir.exists()
    assert list(data_dir.parent.glob("memo-data-backup-*.db"))


def test_reset_and_backup_do_not_create_missing_database(data_dir, capsys):
    code, out, err = run_cli(["reset", "--force"], capsys)
    assert code == 0
    assert err == ""
    assert out.strip() == f"data dir not found: {data_dir}"
    assert not data_dir.exists()

    code, out, err = run_cli(["backup"], capsys)
    assert code == 1
    assert out == ""
    assert f"database not found: {db_path(data_dir)}" in err
    assert not data_dir.exists()


def test_explicit_memo_db_path(data_dir, tmp_path, monkeypatch, capsys):
    explicit_db = tmp_path / "custom" / "memo.db"
    explicit_db.parent.mkdir()
    monkeypatch.setenv("MEMO_DB_PATH", str(explicit_db))

    code, out, err = run_cli(["add", "custom db"], capsys)
    assert code == 0
    assert err == ""
    assert "created #1" in out
    assert explicit_db.exists()
    assert not db_path(data_dir).exists()


def test_backup_export_import_and_replace(data_dir, tmp_path, capsys):
    run_cli(["add", "alpha", "#one"], capsys)
    run_cli(["add", "beta", "#two"], capsys)
    run_cli(["link", "1", "2", "--note", "pair"], capsys)

    backup_path = tmp_path / "backup.db"
    code, out, err = run_cli(["backup", "--out", str(backup_path)], capsys)
    assert code == 0
    assert err == ""
    assert out.strip() == f"backup saved: {backup_path}"
    assert backup_path.exists()

    export_path = tmp_path / "export.json"
    code, out, err = run_cli(["export", "--out", str(export_path)], capsys)
    assert code == 0
    assert err == ""
    assert out.strip() == f"exported: {export_path}"
    payload = json.loads(export_path.read_text(encoding="utf-8"))
    assert payload["format"] == "a-memo-export"
    assert [memo["content"] for memo in payload["memos"]] == ["alpha", "beta"]
    assert payload["memo_links"][0]["note"] == "pair"

    run_cli(["delete", "1"], capsys)
    code, out, err = run_cli(["import", str(export_path), "--replace", "--no-backup"], capsys)
    assert code == 0
    assert err == ""
    assert out.strip() == "imported 2 memos"

    code, out, _ = run_cli(["list"], capsys)
    rows = parse_json_output(out)
    assert {row["id"] for row in rows} == {1, 2}

    code, out, _ = run_cli(["links", "1"], capsys)
    links = parse_json_output(out)
    assert links[0]["id"] == 2
    assert links[0]["note"] == "pair"


def test_import_append_remaps_links(data_dir, tmp_path, capsys):
    run_cli(["add", "existing"], capsys)

    export_path = tmp_path / "append.json"
    export_path.write_text(
        json.dumps(
            {
                "format": "a-memo-export",
                "schema_version": 1,
                "exported_at": time.time(),
                "memos": [
                    {
                        "id": 10,
                        "content": "imported one",
                        "tags": "[]",
                        "created_at": time.time(),
                        "updated_at": None,
                        "review_count": 0,
                        "last_review_at": None,
                        "source": "test",
                    },
                    {
                        "id": 20,
                        "content": "imported two",
                        "tags": "[]",
                        "created_at": time.time(),
                        "updated_at": None,
                        "review_count": 0,
                        "last_review_at": None,
                        "source": "test",
                    },
                ],
                "memo_links": [
                    {
                        "id": 99,
                        "from_memo_id": 10,
                        "to_memo_id": 20,
                        "relation_type": "related",
                        "note": "remapped",
                        "created_at": time.time(),
                        "source": "test",
                    }
                ],
            }
        ),
        encoding="utf-8",
    )

    code, out, err = run_cli(["import", str(export_path)], capsys)
    assert code == 0
    assert err == ""
    assert out.strip() == "imported 2 memos"

    code, out, _ = run_cli(["search", "imported"], capsys)
    rows = parse_json_output(out)
    imported_ids = {row["id"] for row in rows}
    assert imported_ids == {2, 3}

    code, out, _ = run_cli(["links", "2"], capsys)
    links = parse_json_output(out)
    assert links[0]["id"] == 3
    assert links[0]["note"] == "remapped"


def test_legacy_database_migrates_to_current_schema(data_dir, capsys):
    data_dir.mkdir()
    with sqlite3.connect(db_path(data_dir)) as conn:
        conn.executescript(
            """
            CREATE TABLE memos (
              id INTEGER PRIMARY KEY AUTOINCREMENT,
              content TEXT NOT NULL,
              tags TEXT,
              created_at REAL NOT NULL,
              updated_at REAL,
              review_count INTEGER DEFAULT 0,
              last_review_at REAL,
              source TEXT DEFAULT 'cli'
            );
            CREATE TABLE memo_links (
              id INTEGER PRIMARY KEY AUTOINCREMENT,
              from_memo_id INTEGER NOT NULL,
              to_memo_id INTEGER NOT NULL,
              relation_type TEXT DEFAULT 'related',
              note TEXT,
              created_at REAL NOT NULL,
              source TEXT DEFAULT 'agent',
              UNIQUE(from_memo_id, to_memo_id, relation_type)
            );
            INSERT INTO memos (content, tags, created_at, source)
            VALUES ('legacy memo', '[]', 1, 'test');
            """
        )

    code, out, err = run_cli(["search", "legacy"], capsys)
    assert code == 0
    assert err == ""
    rows = parse_json_output(out)
    assert rows[0]["content"] == "legacy memo"
    assert schema_version(data_dir) == 1


def test_import_rejects_bad_payload(data_dir, tmp_path, capsys):
    bad_path = tmp_path / "bad.json"
    bad_path.write_text("{}", encoding="utf-8")

    code, out, err = run_cli(["import", str(bad_path)], capsys)
    assert code == 1
    assert out == ""
    assert "unsupported import format" in err



"""links 子命令"""

import argparse
import json

from .. import linked_memos_for, require_memos


def add_parser(sub):
    p = sub.add_parser("links", help="查看 memo 关联")
    p.add_argument("id", type=positive_int)
    p.add_argument("--limit", type=positive_int, default=20)
    p.set_defaults(func=cmd_links)
    return p


def positive_int(value):
    try:
        parsed = int(value)
    except ValueError as exc:
        raise argparse.ArgumentTypeError("must be a non-negative integer") from exc
    if parsed < 0:
        raise argparse.ArgumentTypeError("must be >= 0")
    return parsed


def cmd_links(conn, args):
    require_memos(conn, [args.id])
    rows = linked_memos_for(conn, args.id, args.limit)
    print(json.dumps(rows, ensure_ascii=False, indent=2))



"""tag 子命令"""

import argparse
import json
import time

from .. import parse_tags, render_memo_text


def add_parser(sub):
    p = sub.add_parser("tag", help="给 memo 加标签")
    p.add_argument("id", type=positive_int)
    p.add_argument("tags", nargs="+")
    p.set_defaults(func=cmd_tag)
    return p


def positive_int(value):
    try:
        parsed = int(value)
    except ValueError as exc:
        raise argparse.ArgumentTypeError("must be a non-negative integer") from exc
    if parsed < 0:
        raise argparse.ArgumentTypeError("must be >= 0")
    return parsed


def cmd_tag(conn, args):
    row = conn.execute("SELECT content, tags FROM memos WHERE id=?", (args.id,)).fetchone()
    if not row:
        raise ValueError(f"memo #{args.id} not found")

    new = parse_tags(" ".join(args.tags))
    if not new:
        raise ValueError("at least one #tag required")

    existing = json.loads(row["tags"]) if row["tags"] else []
    merged = list(dict.fromkeys(existing + new))

    with conn:
        conn.execute(
            "UPDATE memos SET tags=?, updated_at=? WHERE id=?",
            (json.dumps(merged, ensure_ascii=False), time.time(), args.id),
        )
        conn.execute(
            "INSERT OR REPLACE INTO memos_fts(rowid, content) VALUES (?, ?)",
            (args.id, render_memo_text(merged, row["content"])),
        )
    print(f"tagged #{args.id}: {' '.join('#' + t.lstrip('#') for t in merged)}")



"""review 子命令"""

import argparse
import json
import os
import time

from .. import (
    choose_review_layer,
    fmt_history_time,
    fmt_time,
    get_history_dir,
    history_file_path,
    linked_memos_for,
    render_memo_text,
)


def add_parser(sub):
    p = sub.add_parser("review", help="回顾 memos")
    p.add_argument("--count", type=positive_int, default=5)
    p.add_argument("--push", action="store_true")
    p.set_defaults(func=cmd_review)
    return p


def positive_int(value):
    try:
        parsed = int(value)
    except ValueError as exc:
        raise argparse.ArgumentTypeError("must be a non-negative integer") from exc
    if parsed < 0:
        raise argparse.ArgumentTypeError("must be >= 0")
    return parsed


def select_review_row_from_layer(conn, layer):
    min_days, max_days = _layer_bounds(layer)
    max_filter = "" if max_days is None else "AND days_since_review < ?"
    params = [min_days]
    if max_days is not None:
        params.append(max_days)
    return conn.execute(
        f"""
        WITH eligible AS (
          SELECT
            *,
            COALESCE(last_review_at, created_at) as sort_time,
            ((unixepoch() - COALESCE(last_review_at, created_at)) / 86400.0) as days_since_review
          FROM memos
        ),
        weighted AS (
          SELECT *,
            (days_since_review / (review_count + 1)) *
            (0.7 + abs(random() / 9223372036854775808.0) * 0.6) as score
          FROM eligible
          WHERE days_since_review >= ?
          {max_filter}
        )
        SELECT id, content, tags, created_at, updated_at, review_count
        FROM weighted ORDER BY score DESC LIMIT 1
        """,
        params,
    ).fetchone()


def _layer_bounds(layer):
    if layer == "fresh":
        return 7, 30
    if layer == "middle":
        return 30, 180
    if layer == "old":
        return 180, None
    raise ValueError(f"unknown review layer: {layer}")


def _fallback_layers(layer):
    order = ("fresh", "middle", "old")
    idx = order.index(layer) if layer in order else -1
    return (*order[idx + 1 :], *order[: idx + 1])


def select_single_review_row(conn):
    selected = choose_review_layer()
    for layer in (selected, *_fallback_layers(selected)):
        row = select_review_row_from_layer(conn, layer)
        if row:
            return row
    return None


def cmd_review(conn, args):
    if args.count == 1:
        row = select_single_review_row(conn)
        rows = [row] if row else []
    else:
        rows = conn.execute(
            """
          WITH eligible AS (
            SELECT *, COALESCE(last_review_at, created_at) as sort_time
            FROM memos
            WHERE COALESCE(last_review_at, created_at) < (unixepoch() - 604800)
          ),
          weighted AS (
            SELECT *,
              (1.0 / (review_count + 1)) *
              ((unixepoch() - sort_time) / 86400.0) as weight
            FROM eligible
          )
          SELECT id, content, tags, created_at, updated_at, review_count
          FROM weighted ORDER BY weight DESC LIMIT ?
        """,
            (args.count,),
        ).fetchall()

    if not rows:
        print("[]")
        return

    ids = [row["id"] for row in rows]

    result = []
    for row in rows:
        tags = json.loads(row["tags"]) if row["tags"] else []
        result.append(
            {
                "id": row["id"],
                "content": row["content"].strip(),
                "tags": tags,
                "created_at": fmt_time(row["created_at"]) if row["created_at"] else None,
                "updated_at": fmt_time(row["updated_at"]) if row["updated_at"] else None,
                "review_count": row["review_count"] or 0,
                "links": linked_memos_for(conn, row["id"]),
            }
        )

    print(json.dumps(result, ensure_ascii=False, indent=2))

    if args.push and ids:
        now = time.time()
        _append_history(rows, now)
        placeholders = ",".join("?" * len(ids))
        with conn:
            conn.execute(
                f"""
                UPDATE memos
                SET review_count=review_count+1, last_review_at=?
                WHERE id IN ({placeholders})
                """,
                [now] + ids,
            )


def _append_history(rows, reviewed_at=None):
    if not rows:
        return
    reviewed_at = reviewed_at or time.time()
    history_dir = get_history_dir()
    history_dir.mkdir(parents=True, exist_ok=True)
    out = history_file_path(reviewed_at, history_dir)

    entries = []
    for row in rows:
        tags = json.loads(row["tags"]) if row["tags"] else []
        memo = render_memo_text(tags, row["content"].strip())
        entries.append(f"{fmt_history_time(reviewed_at)}\n{memo}\n---")

    prefix = ""
    if out.exists() and out.stat().st_size > 0:
        with out.open("rb") as fh:
            fh.seek(-1, os.SEEK_END)
            if fh.read(1) != b"\n":
                prefix = "\n"

    with out.open("a", encoding="utf-8") as fh:
        fh.write(prefix + "\n".join(entries) + "\n")



"""unlink 子命令"""

import argparse

from .. import normalize_link_ids, validate_relation_type


def add_parser(sub):
    p = sub.add_parser("unlink", help="删除两条 memo 的关联")
    p.add_argument("left_id", type=positive_int)
    p.add_argument("right_id", type=positive_int)
    p.add_argument("--type", default="related", help="关联类型: related, supports, contrasts")
    p.set_defaults(func=cmd_unlink)
    return p


def positive_int(value):
    try:
        parsed = int(value)
    except ValueError as exc:
        raise argparse.ArgumentTypeError("must be a non-negative integer") from exc
    if parsed < 0:
        raise argparse.ArgumentTypeError("must be >= 0")
    return parsed


def cmd_unlink(conn, args):
    relation_type = validate_relation_type(args.type)
    from_memo_id, to_memo_id = normalize_link_ids(args.left_id, args.right_id)

    try:
        with conn:
            cur = conn.execute(
                """
                DELETE FROM memo_links
                WHERE from_memo_id=? AND to_memo_id=? AND relation_type=?
                """,
                (from_memo_id, to_memo_id, relation_type),
            )
    except Exception as exc:
        raise RuntimeError(f"failed to unlink: {exc}") from exc

    if cur.rowcount == 0:
        raise ValueError(f"#{args.left_id} ↔ #{args.right_id} has no {relation_type} link")
    print(f"unlinked #{args.left_id} ↔ #{args.right_id} ({relation_type})")



"""memo CLI 入口"""

import argparse
import sys

from . import connect, get_db_path


def _get_version():
    try:
        from importlib.metadata import version

        return version("a-memo")
    except Exception:
        return "0.1.0"


def positive_int(value):
    try:
        parsed = int(value)
    except ValueError as exc:
        raise argparse.ArgumentTypeError("must be a non-negative integer") from exc
    if parsed < 0:
        raise argparse.ArgumentTypeError("must be >= 0")
    return parsed


COMMANDS = [
    "init",
    "reset",
    "add",
    "update",
    "delete",
    "list",
    "search",
    "review",
    "tag",
    "tags",
    "link",
    "unlink",
    "links",
    "backup",
    "export",
    "import",
    "rebuild-fts",
    "image",
    "flomo-import",
]

_MOD_ALIASES = {
    "rebuild-fts": "rebuild_fts",
    "flomo-import": "flomo_import",
    "import": "import_data",
}

_command_modules = {}


def _load(name):
    mod_name = _MOD_ALIASES.get(name, name)
    if mod_name not in _command_modules:
        try:
            mod = __import__(f"memo.commands.{mod_name}", fromlist=["add_parser"])
            _command_modules[mod_name] = mod
        except ImportError:
            _command_modules[mod_name] = None
    return _command_modules.get(mod_name)


def _levenshtein(a, b):
    """计算两个字符串之间的编辑距离"""
    if len(a) < len(b):
        return _levenshtein(b, a)
    if len(b) == 0:
        return len(a)

    previous_row = range(len(b) + 1)
    for i, ca in enumerate(a):
        current_row = [i + 1]
        for j, cb in enumerate(b):
            insertions = previous_row[j + 1] + 1
            deletions = current_row[j] + 1
            substitutions = previous_row[j] + (ca != cb)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row

    return previous_row[-1]


def _suggest(cmd):
    """找到最相似的命令"""
    candidates = COMMANDS + list(_MOD_ALIASES.keys())
    best, best_dist = None, float("inf")
    for c in candidates:
        d = _levenshtein(cmd, c)
        if d < best_dist:
            best, best_dist = c, d
    return best, best_dist


def _print_help_and_exit(exit_code=0):
    print("memo — lightweight memo CLI for AI agents")
    print()
    print("USAGE:")
    print("    memo <command> [options]")
    print()
    print("COMMANDS:")
    print("    init           initialize database")
    print("    reset          reset data directory")
    print("    add            add memo")
    print("    list           list memos")
    print("    search         search memos")
    print("    review         review memos")
    print("    update         update memo")
    print("    delete         delete memo")
    print("    tag            tag memo")
    print("    tags           list all tags")
    print("    link           link two memos")
    print("    unlink         unlink two memos")
    print("    links          view memo links")
    print("    backup         backup SQLite database")
    print("    export         export data as JSON")
    print("    import         import JSON export")
    print("    image          generate share image")
    print("    rebuild-fts    rebuild search index")
    print("    flomo-import   import from flomo HTML")
    print()
    print("FLAGS:")
    print("    -h, --help     show help")
    print("    -v, --version  show version")
    print()
    print("more: memo <command> --help")
    sys.exit(exit_code)


def _print_version_and_exit():
    print(f"memo {_get_version()}")
    sys.exit(0)


def main(argv=None):
    raw = argv if argv is not None else sys.argv[1:]

    if not raw or raw[0] in ("-h", "--help"):
        _print_help_and_exit()

    if raw[0] in ("-v", "--version"):
        _print_version_and_exit()

    parser = argparse.ArgumentParser(prog="memo", add_help=False, exit_on_error=False)
    sub = parser.add_subparsers(dest="cmd", required=True)

    for cmd in COMMANDS:
        mod = _load(cmd)
        if mod:
            mod.add_parser(sub)

    try:
        args = parser.parse_args(raw)
    except SystemExit:
        # argparse --help exits, let it through
        raise
    except BaseException:
        # Catch invalid command and suggest similar one
        if raw and raw[0] not in COMMANDS:
            suggestion, dist = _suggest(raw[0])
            print(f"error: unknown command '{raw[0]}'", file=sys.stderr)
            if dist <= 3:
                print(f"       did you mean '{suggestion}'?", file=sys.stderr)
            print()
            _print_help_and_exit(2)
        raise

    if args.cmd == "reset":
        try:
            return args.func(None, args) or 0
        except (RuntimeError, ValueError) as exc:
            print(str(exc), file=sys.stderr)
            return 1

    if args.cmd == "backup" and not get_db_path().exists():
        print(f"database not found: {get_db_path()}", file=sys.stderr)
        return 1

    conn = connect()
    try:
        return args.func(conn, args) or 0
    except (RuntimeError, ValueError) as exc:
        print(str(exc), file=sys.stderr)
        return 1
    finally:
        conn.close()


if __name__ == "__main__":
    sys.exit(main())
